//
// Created by philipp on 02.01.18.
//

#ifndef NETWORKINGLIB_CLOSEABLE_H_H
#define NETWORKINGLIB_CLOSEABLE_H_H

#include "Networking.h"
#include "Timer.h"
#include "Error.h"

namespace networking
{
namespace closeable
{

template<typename Closeable>
class Closer
{
public:
    Closer(Closeable & stream) : stream(stream)
    {}

    ~Closer()
    {
        close(stream);
    }

    static void close(Closeable & socket)
    {
        boost::system::error_code ignoredError;
        socket.close(ignoredError);
    }

private:
    Closeable & stream;
};

template<typename AsyncOperation, typename... AsyncOperationArgs, typename Closeable>
void timedOperation(Networking & net,
                    AsyncOperation asyncOperation,
                    Closeable & closeable,
                    const time::Duration & timeout,
                    AsyncOperationArgs && ... args)
{
    auto & ioService = net.getIoService();

    auto timer = time::Timer::create(net);
    timer->startTimeout(
        timeout,
        [&closeable]
        {
            // Timeout expired: close the closeable.
            Closer<Closeable>::close(closeable);
        });

    // A 'would_block' closeableError is guaranteed to never occur on an asynchronous operation.
    boost::system::error_code closeableError = boost::asio::error::would_block;

    // Run asynchronous connect.
    asyncOperation(
        std::forward<AsyncOperationArgs>(args)...,
        [&closeableError, timer](const boost::system::error_code & error, auto ...)
        {
            timer->stop();
            // update closeableError variable
            closeableError = error;
        });

    // Wait until "something happens" with the socket.
    net.waitUntil([&closeableError]()
                  { return closeableError != boost::asio::error::would_block; });

    // Determine whether a connection was successfully established.
    // Even though our timer handler might have run to close the closeable, the connect operation
    // might have notionally succeeded!
    if (closeableError || !closeable.is_open())
    {
        if (closeableError == boost::asio::error::operation_aborted)
            throw error::Aborted{};

        throw error::FailedOperation{};
    }
}

template<
    typename AsyncOperation,
    typename... AsyncOperationArgs,
    typename Closeable,
    typename Handler>
void timedAsyncOperation(Networking & net,
                         AsyncOperation asyncOperation,
                         Closeable & closeable,
                         const time::Duration & timeout,
                         Handler handler,
                         AsyncOperationArgs && ... asyncOperationArgs)
{
    auto timer = time::Timer::create(net);
    timer->startTimeout(
        timeout,
        [&closeable]
        {
            Closer<Closeable>::close(closeable);
        });

    asyncOperation(
        std::forward<AsyncOperationArgs>(asyncOperationArgs)...,
        [&closeable, timer, handler](const boost::system::error_code & opError, auto && ... handlerArgs)
        {
            timer->stop();

            auto errorCode = error::codes::SUCCESS;
            if (!closeable.is_open())
                errorCode = error::codes::ABORTED;
            else if (opError)
                errorCode = error::codes::FAILED_OPERATION;

            handler(errorCode, std::forward<decltype(handlerArgs)>(handlerArgs)...);
        });
}

}
}

#endif //NETWORKINGLIB_CLOSEABLE_H_H
