//
// Created by philipp on 29.12.17.
//

#ifndef NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H
#define NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H

#include "Networking.h"
#include "Stream.h"
#include "Socket.h"
#include "Message.h"

namespace networking
{
namespace message
{

template<typename Message>
class DatagramReceiver : public std::enable_shared_from_this<DatagramReceiver<Message>>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<DatagramReceiver<Message>>;

    using Buffer = typename Message::Buffer;

    using Udp = boost::asio::ip::udp;
    using Endpoint = Udp::endpoint;
    using Socket = Udp::socket;

    using ReceiveHandler = std::function<
        void(const error::ErrorCode & error,
             const Message & message,
             const std::string & senderHost,
             std::uint16_t senderPort)>;

    DatagramReceiver(PrivateTag, Networking & net, std::uint16_t bindingPort)
        : net(net)
          , bindingPort(bindingPort)
          , socket(net.getIoService(), Udp::v4())
          , buffer(Message::emptyBuffer())
    {}

    static Ptr create(Networking & net, std::uint16_t bindingPort)
    {
        return std::make_shared<DatagramReceiver<Message>>(PrivateTag{}, net, bindingPort);
    }

    bool isReceiving() const noexcept
    { return receiving; };

    void asyncReceive(const time::Duration & timeout, ReceiveHandler receiveHandler)
    {
        if (receiving)
            throw error::Busy{};

        try
        {
            receiving = true;
            newSocket();

            auto self = this->shared_from_this();
            message::asyncReceiveDatagramFrom<Message>(
                net, socket, timeout,
                [self, receiveHandler](const auto & error,
                       const auto & message,
                       const auto & senderHost,
                       auto senderPort)
                {
                    self->finishReceiving();
                    receiveHandler(error, message, senderHost, senderPort);
                });
        }
        catch (...)
        {
            finishReceiving();
            throw error::FailedOperation{};
        }
    }

    void stop()
    {
        if (!receiving)
            return;

        closeable::Closer<Socket>::close(socket);
    }

private:
    Networking & net;
    std::uint16_t bindingPort;
    Socket socket;
    Buffer buffer;
    std::atomic<bool> receiving{false};

    void newSocket()
    {
        socket = Socket(net.getIoService(), Udp::v4());
        socket.set_option(boost::asio::socket_base::reuse_address{true});
        socket.bind(Endpoint(Udp::v4(), bindingPort));
    }

    void finishReceiving()
    {
        receiving = false;
        closeable::Closer<Socket>::close(socket);
    }
};

}
}

#endif //NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H
