//
// Created by philipp on 29.12.17.
//

#ifndef NETWORKINGLIB_DATAGRAMSENDER_H
#define NETWORKINGLIB_DATAGRAMSENDER_H

#include "Stream.h"
#include "Networking.h"
#include "Message.h"
#include "Utils.h"

namespace networking
{
namespace message
{

template<typename Message>
class DatagramSender : public std::enable_shared_from_this<DatagramSender<Message>>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<DatagramSender>;

    using SendHandler = std::function<void(const error::ErrorCode & error)>;

    using Udp = boost::asio::ip::udp;
    using Socket = Udp::socket;

    static Ptr create(Networking & net)
    {
        return std::make_shared<DatagramSender>(PrivateTag{}, net);
    }

    DatagramSender(PrivateTag, Networking & net)
        : net(net)
          , socket(net.getIoService())
    {}

    void send(const Message & message,
              const std::string & host,
              std::uint16_t port,
              const time::Duration & timeout)
    {
        if (sending)
            throw error::Busy();

        // Use RAII to reset the sending flag when leaving this function.
        utils::RAIIObject sendingFlagReset{[this]
                                           { sending = false; }};
        openSocket();
        networking::message::sendDatagramTo(net, socket, message, host, port, timeout);
    }

    void asyncSend(const Message & message,
                   const std::string & host,
                   std::uint16_t port,
                   const time::Duration & timeout,
                   const SendHandler & sendHandler)
    {
        if (sending)
            throw error::Busy{};

        try
        {
            sending = true;
            openSocket();

            auto self = this->shared_from_this();
            networking::message::asyncSendDatagramTo(
                net, socket, message, host, port, timeout,
                [self, sendHandler](const auto & error)
                {
                    self->finishSending();
                    sendHandler(error);
                });
        }
        catch (...)
        {
            finishSending();
            throw error::FailedOperation{};
        }
    }

    bool isSending() const noexcept
    {
        return sending;
    }

    void stop()
    {
        if (!sending)
            return;

        closeable::Closer<Socket>::close(socket);
    }

private:
    Networking & net;
    Socket socket;
    std::atomic<bool> sending{false};

    void openSocket()
    {
        if (!socket.is_open())
        {
            boost::system::error_code ignoredError;
            socket.open(Udp::v4(), ignoredError);
        }
    }

    void finishSending()
    {
        sending = false;
    }
};

}
}

#endif //NETWORKINGLIB_DATAGRAMSENDER_H
