//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_MESSAGES_H
#define PROTOCOL_MESSAGES_H

#include <cstdint>
#include <vector>
#include <boost/system/error_code.hpp>
#include <boost/asio/write.hpp>
#include <boost/asio/read.hpp>
#include "Stream.h"
#include "Networking.h"
#include "Socket.h"

namespace networking
{
namespace message
{

using SendHandler = std::function<void(const error::ErrorCode & code)>;

template<typename Message>
using ReceiveHandler = std::function<void(const error::ErrorCode & code, const Message & message)>;

using SendToHandler = std::function<void(const error::ErrorCode & code)>;

template<typename Message>
using ReceiveFromHandler = std::function<
    void(const error::ErrorCode & code,
         const Message & message,
         const std::string & senderHost,
         std::uint16_t senderPort)>;

template<typename Message>
void send(Networking & net,
          boost::asio::ip::tcp::socket & socket,
          const Message & message,
          const time::Duration & timeout)
{
    using namespace std::chrono_literals;
    if (timeout <= 0s)
        throw error::Aborted();

    networking::stream::write(net, socket, boost::asio::buffer(message.data()), timeout);
};

template<typename Message>
void asyncSend(Networking & net,
               boost::asio::ip::tcp::socket & socket,
               const Message & message,
               const time::Duration & timeout,
               SendHandler handler)
{
    networking::stream::asyncWrite(
        net, socket, boost::asio::buffer(message.data()), timeout,
        [handler](const auto & errorCode, auto ...)
        {
            handler(errorCode);
        });
};

template<typename Message>
Message receive(Networking & net,
                boost::asio::ip::tcp::socket & socket,
                const time::Duration & timeout)
{
    using namespace std::chrono_literals;
    if (timeout <= 0s)
        throw error::Aborted();

    auto buffer = Message::emptyBuffer();
    networking::stream::read(net, socket, boost::asio::buffer(buffer), timeout);
    return Message(buffer);
};

template<typename Message>
void asyncReceive(Networking & net,
                  boost::asio::ip::tcp::socket & socket,
                  const time::Duration & timeout,
                  ReceiveHandler<Message> handler)
{
    auto buffer = std::make_shared<typename Message::Buffer>(Message::emptyBuffer());
    networking::stream::asyncRead(
        net, socket, boost::asio::buffer(*buffer), timeout,
        [buffer, handler](const auto & errorCode, auto ...)
        {
            handler(errorCode, Message(*buffer));
        });
};

template<typename Message>
void sendDatagramTo(Networking & net,
                    boost::asio::ip::udp::socket & socket,
                    const Message & message,
                    const std::string & host,
                    std::uint16_t port,
                    const time::Duration & timeout)
{
    networking::socket::sendTo(net, socket, boost::asio::buffer(message.data()), host, port, timeout);
}

template<typename Message>
void asyncSendDatagramTo(Networking & net,
                         boost::asio::ip::udp::socket & socket,
                         const Message & message,
                         const std::string & host,
                         std::uint16_t port,
                         const time::Duration & timeout,
                         SendToHandler handler)
{
    networking::socket::asyncSendTo(
        net, socket, boost::asio::buffer(message.data()), host, port, timeout,
        [handler](const auto & error, auto ...)
        {
            handler(error);
        });
}

template<typename Message>
Message receiveDatagramFrom(Networking & net,
                            boost::asio::ip::udp::socket & socket,
                            std::string & host,
                            std::uint16_t & port,
                            const time::Duration & timeout)
{
    auto data = Message::emptyBuffer();
    networking::socket::receiveFrom(net, socket, boost::asio::buffer(data), host, port, timeout);
    return Message(data);
}

template<typename Message>
void asyncReceiveDatagramFrom(Networking & net,
                              boost::asio::ip::udp::socket & socket,
                              const time::Duration & timeout,
                              ReceiveFromHandler<Message> handler)
{
    auto buffer = std::make_shared<typename Message::Buffer>(Message::emptyBuffer());
    networking::socket::asyncReceiveFrom(
        net, socket, boost::asio::buffer(*buffer), timeout,
        [buffer, handler](auto error, auto bytesTransferred, const auto & senderHost, auto senderPort)
        {
            Message msg{*buffer};

            if (!error && bytesTransferred < buffer->size())
                error = networking::error::codes::FAILED_OPERATION;

            handler(error, msg, senderHost, senderPort);
        });
}

}
}


#endif //PROTOCOL_MESSAGES_H
