//
// Created by philipp on 02.01.18.
//

#ifndef NETWORKINGLIB_SOCKETOPS_H
#define NETWORKINGLIB_SOCKETOPS_H

#include "Stream.h"

namespace networking
{
namespace socket
{

template<typename SocketService>
void connect(Networking & net,
             SocketService & socket,
             const std::string & host,
             std::uint16_t port,
             const time::Duration & timeout)
{
    boost::asio::ip::tcp::resolver resolver(net.getIoService());
    auto endpointIterator = resolver.resolve({host, std::to_string(port)});
    // This lambda is used because we cannot pass boost::asio::async_connect directly as a function
    // since it is a template.
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };
    closeable::timedOperation(net, asyncOperation, socket, timeout, socket, endpointIterator);
}

template<typename SocketService, typename ConnectHandler>
void asyncConnect(Networking & net,
                  SocketService & socket,
                  const std::string & host,
                  std::uint16_t port,
                  const time::Duration & timeout,
                  ConnectHandler handler)
{
    boost::asio::ip::tcp::resolver resolver(net.getIoService());
    auto endpointIterator = resolver.resolve({host, std::to_string(port)});
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };
    closeable::timedAsyncOperation(net, asyncOperation, socket, timeout, handler, socket, endpointIterator);
}

template<
    typename DatagramSocket,
    typename MutableBuffers>
void receiveFrom(Networking & net,
                 DatagramSocket & socket,
                 const MutableBuffers & buffer,
                 std::string & senderHost,
                 std::uint16_t & senderPort,
                 const time::Duration & timeout)
{
    boost::asio::ip::udp::endpoint senderEndpoint;
    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_receive_from(std::forward<decltype(args)>(args)...); };
    closeable::timedOperation(net, asyncOperation, socket, timeout, buffer, senderEndpoint);
    senderHost = senderEndpoint.address().to_string();
    senderPort = senderEndpoint.port();
}

template<
    typename DatagramSocket,
    typename MutableBuffers,
    typename ReceiveHandler>
void asyncReceiveFrom(Networking & net,
                      DatagramSocket & socket,
                      const MutableBuffers & buffer,
                      const time::Duration & timeout,
                      ReceiveHandler handler)
{
    auto senderEndpoint = std::make_shared<boost::asio::ip::udp::endpoint>();
    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_receive_from(std::forward<decltype(args)>(args)...); };
    closeable::timedAsyncOperation(
        net, asyncOperation, socket, timeout,
        [handler, senderEndpoint](auto && ... args)
        {
            auto senderHost = senderEndpoint->address().to_string();
            auto senderPort = senderEndpoint->port();
            handler(std::forward<decltype(args)>(args)..., senderHost, senderPort);
        }, buffer, *senderEndpoint);
}

template<
    typename DatagramSocket,
    typename MutableBuffers>
void sendTo(Networking & net,
            DatagramSocket & socket,
            const MutableBuffers & buffer,
            const std::string & host,
            std::uint16_t port,
            const time::Duration & timeout)
{
    boost::asio::ip::udp::endpoint endpoint{boost::asio::ip::address::from_string(host), port};
    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_send_to(std::forward<decltype(args)>(args)...); };
    closeable::timedOperation(net, asyncOperation, socket, timeout, buffer, endpoint);
};

template<
    typename DatagramSocket,
    typename MutableBuffers,
    typename SendHandler>
void asyncSendTo(Networking & net,
                 DatagramSocket & socket,
                 const MutableBuffers & buffer,
                 const std::string & host,
                 std::uint16_t port,
                 const time::Duration & timeout,
                 SendHandler handler)
{
    boost::asio::ip::udp::endpoint endpoint{boost::asio::ip::address::from_string(host), port};
    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_send_to(std::forward<decltype(args)>(args)...); };
    closeable::timedAsyncOperation(net, asyncOperation, socket, timeout, handler, buffer, endpoint);
};

}
}

#endif //NETWORKINGLIB_SOCKETOPS_H
