//
// Created by philipp on 30.11.17.
//

#include <vector>
#include <algorithm>
#include <iostream>
#include "../include/Vehicle.h"
#include "../include/Utils.h"

using namespace protocol;

Vehicle::Vehicle(networking::Networking & net, VehicleId vehicleId, Vehicle::Role role)
    : net(net)
      , vehicleId(vehicleId)
      , role(role)
{
    platoonCreateServiceServer = networking::service::Server<PlatoonService>::create(net,
                                                                                     getOwnEndpoint().getTcpPort());
    platoonCreateServiceClient = networking::service::Client<PlatoonService>::create(net);
}

bool Vehicle::isIdle()
{
    std::lock_guard<std::mutex> lock(stateMutex);
    return state == State::IDLE;
}

bool Vehicle::isPlatoonRunning()
{
    std::lock_guard<std::mutex> lock(stateMutex);
    return state == State::RUNNING_PLATOON;
}

void Vehicle::createPlatoon()
{
    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->doCreatePlatoon(); });
}

void Vehicle::leavePlatoon()
{
    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->doLeavePlatoon(); });
}

void Vehicle::stop()
{
    platoonCreateServiceServer->stop();

    platoonCreateServiceClient->stop();
}

void Vehicle::doCreatePlatoon()
{
    auto self = shared_from_this();
    platoonCreateServiceServer->advertiseService(
        [self](const auto & clientEndpoint, const auto & requestMessage) -> PlatoonMessage
        { return self->makePlatoonCreateResponse(clientEndpoint, requestMessage); });

    // We cycle through every available endpoint and try to make a create-platoon-request.
    scanAvailableVehicles();
    vehicleEndpointIter = availableVehicleEndpoints.begin();
    makePlatoonCreateRequestToNextEndpoint();
}

void Vehicle::makePlatoonCreateRequestToNextEndpoint()
{
    if (state != State::CREATING_PLATOON)
        return;

    if (!platoonCreateServiceClient->isCalling())
    {
        auto vehicleEndpoint = *vehicleEndpointIter;
        makePlatoonCreateRequest(vehicleEndpoint);
        utils::cycle(vehicleEndpointIter, availableVehicleEndpoints);
    }

    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->makePlatoonCreateRequestToNextEndpoint(); });
}

void Vehicle::scanAvailableVehicles()
{
    // TODO: change vehicle endpoint information!
    availableVehicleEndpoints = {
        VehicleEndpoint{1, "127.0.0.1", 10001, 10011},
        VehicleEndpoint{2, "127.0.0.1", 10002, 10022},
        VehicleEndpoint{3, "127.0.0.1", 10003, 10033},
    };

    // remove me out of the list (I cannot connect to myself)
    availableVehicleEndpoints.erase(std::remove_if(
        availableVehicleEndpoints.begin(), availableVehicleEndpoints.end(),
        [this](VehicleEndpoint & endpoint) -> bool
        {
            return endpoint.getVehicleId() == vehicleId;
        }), availableVehicleEndpoints.end());
}

void Vehicle::setState(Vehicle::State newState)
{
    std::lock_guard<std::mutex> lock(stateMutex);
    state = newState;
}

VehicleEndpoint Vehicle::getOwnEndpoint() const
{
    // TODO: How to get this in real usage???
    return VehicleEndpoint{
        vehicleId,
        "127.0.0.1",
        vehicleId + 10000,
        vehicleId + 10010};
}
